/*
 * CzDC - client for Direct Connect peer to peer network.

 * Copyright (C) 2002-2011 Petr Kozelka, PPK at CzDC dot org

 * CzDC is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 3
 * as published by the Free Software Foundation.

 * CzDC is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.

 * You should have received a copy of the GNU General Public License
 * along with CzDC. If not, see <http://www.gnu.org/licenses/>.
 */

//---------------------------------------------------------------------------
#include "stdafx.h"
//---------------------------------------------------------------------------
#include "../client/DCPlusPlus.h"
//---------------------------------------------------------------------------
#include "../client/SettingsManager.h"
#include "../client/Socket.h"
#include "../client/Util.h"
//---------------------------------------------------------------------------
#include "Resource.h"
//---------------------------------------------------------------------------
#include "IP2Country.h"
//---------------------------------------------------------------------------
IP2Country * g_pIP2Country = NULL;
//---------------------------------------------------------------------------

static const char * CountryNames[] = { "Andorra", "United Arab Emirates", "Afghanistan", "Antigua and Barbuda", "Anguilla", "Albania", "Armenia", "Angola", "Antarctica", "Argentina",
    "American Samoa", "Austria", "Australia", "Aruba", "Aland Islands", "Azerbaijan", "Bosnia and Herzegovina", "Barbados", "Bangladesh", "Belgium", "Burkina Faso", "Bulgaria", "Bahrain",
    "Burundi", "Benin", "Saint Barthelemy", "Bermuda", "Brunei Darussalam", "Bolivia", "Bonaire", "Brazil", "Bahamas", "Bhutan", "Bouvet Island", "Botswana", "Belarus", "Belize", "Canada",
    "Cocos (Keeling) Islands", "The Democratic Republic of the Congo", "Central African Republic", "Congo", "Cote D'Ivoire", "Cook Islands", "Chile", "Cameroon", "China", "Colombia",
    "Costa Rica", "Cuba", "Cape Verde", "Curacao", "Christmas Island", "Cyprus", "Czech Republic", "Germany", "Djibouti", "Denmark", "Dominica", "Dominican Republic", "Algeria", "Ecuador",
    "Estonia", "Egypt", "Western Sahara", "Eritrea", "Spain", "Ethiopia", "Finland", "Fiji", "Falkland Islands (Malvinas)", "Micronesia", "Faroe Islands", "France", "Gabon", "United Kingdom",
    "Grenada", "Georgia", "French Guiana", "Guernsey", "Ghana", "Gibraltar", "Greenland", "Gambia", "Guinea", "Guadeloupe", "Equatorial Guinea", "Greece",
    "South Georgia and the South Sandwich Islands", "Guatemala", "Guam", "Guinea-Bissau", "Guyana", "Hong Kong", "Heard Island and McDonald Islands", "Honduras", "Croatia", "Haiti", "Hungary",
    "Switzerland", "Indonesia", "Ireland", "Israel", "Isle of Man", "India", "British Indian Ocean Territory", "Iraq", "Iran", "Iceland", "Italy", "Jersey", "Jamaica", "Jordan", "Japan",
    "Kenya", "Kyrgyzstan", "Cambodia", "Kiribati", "Comoros", "Saint Kitts and Nevis", "Democratic People's Republic of Korea", "Republic of Korea", "Kuwait", "Cayman Islands", "Kazakhstan",
    "Lao People's Democratic Republic", "Lebanon", "Saint Lucia", "Liechtenstein", "Sri Lanka", "Liberia", "Lesotho", "Lithuania", "Luxembourg", "Latvia", "Libyan Arab Jamahiriya", "Morocco",
    "Monaco", "Moldova", "Montenegro", "Saint Martin", "Madagascar", "Marshall Islands", "Macedonia", "Mali", "Myanmar", "Mongolia", "Macao", "Northern Mariana Islands", "Martinique",
    "Mauritania", "Montserrat", "Malta", "Mauritius", "Maldives", "Malawi", "Mexico", "Malaysia", "Mozambique", "Namibia", "New Caledonia", "Niger", "Norfolk Island", "Nigeria", "Nicaragua",
    "Netherlands", "Norway", "Nepal", "Nauru", "Niue", "New Zealand", "Oman", "Panama", "Peru", "French Polynesia", "Papua New Guinea", "Philippines", "Pakistan", "Poland",
    "Saint Pierre and Miquelon", "Pitcairn", "Puerto Rico", "Palestinian Territory", "Portugal", "Palau", "Paraguay", "Qatar", "Reunion", "Romania", "Serbia", "Russian Federation",
    "Rwanda", "Saudi Arabia", "Solomon Islands", "Seychelles", "Sudan", "Sweden", "Singapore", "Saint Helena", "Slovenia", "Svalbard and Jan Mayen", "Slovakia", "Sierra Leone", "San Marino",
    "Senegal", "Somalia", "Suriname", "South Sudan", "Sao Tome and Principe", "El Salvador", "Sint Maarten (Dutch Part)", "Syrian Arab Republic", "Swaziland", "Turks and Caicos Islands", "Chad",
    "French Southern Territories", "Togo", "Thailand", "Tajikistan", "Tokelau", "Timor-Leste", "Turkmenistan", "Tunisia", "Tonga", "Turkey", "Trinidad and Tobago", "Tuvalu", "Taiwan",
    "Tanzania", "Ukraine", "Uganda", "United States Minor Outlying Islands", "United States", "Uruguay", "Uzbekistan", "Holy See (Vatican City State)",
    "Saint Vincent and the Grenadines", "Venezuela", "Virgin Islands, British", "Virgin Islands, U.S.", "Viet Nam", "Vanuatu", "Wallis and Futuna", "Samoa", "Yemen", "Mayotte", "South Africa",
    "Zambia", "Zimbabwe", "Netherlands Antilles", "Unknown (Asia-Pacific)", "Unknown (European Union)", "Unknown",
};
// last updated 25 sep 2011
static const char * CountryCodes[] = { "AD", "AE", "AF", "AG", "AI", "AL", "AM", "AO", "AQ", "AR",
    "AS", "AT", "AU", "AW", "AX", "AZ", "BA", "BB", "BD", "BE", "BF", "BG", "BH",
    "BI", "BJ", "BL", "BM", "BN", "BO", "BQ", "BR", "BS", "BT", "BV", "BW", "BY", "BZ", "CA",
    "CC", "CD", "CF", "CG", "CI", "CK", "CL", "CM", "CN", "CO",
    "CR", "CU", "CV", "CW", "CX", "CY", "CZ", "DE", "DJ", "DK", "DM", "DO", "DZ", "EC",
    "EE", "EG", "EH", "ER", "ES", "ET", "FI", "FJ", "FK", "FM", "FO", "FR", "GA", "GB",
	"GD", "GE", "GF", "GG", "GH", "GI", "GL", "GM", "GN", "GP", "GQ", "GR",
    "GS", "GT", "GU", "GW", "GY", "HK", "HM", "HN", "HR", "HT", "HU",
    "CH", "ID", "IE", "IL", "IM", "IN", "IO", "IQ", "IR", "IS", "IT", "JE", "JM", "JO", "JP",
    "KE", "KG", "KH", "KI", "KM", "KN", "KP", "KR", "KW", "KY", "KZ",
    "LA", "LB", "LC", "LI", "LK", "LR", "LS", "LT", "LU", "LV", "LY", "MA",
    "MC", "MD", "ME", "MF", "MG", "MH", "MK", "ML", "MM", "MN", "MO", "MP", "MQ",
    "MR", "MS", "MT", "MU", "MV", "MW", "MX", "MY", "MZ", "NA", "NC", "NE", "NF", "NG", "NI",
    "NL", "NO", "NP", "NR", "NU", "NZ", "OM", "PA", "PE", "PF", "PG", "PH", "PK", "PL",
    "PM", "PN", "PR", "PS", "PT", "PW", "PY", "QA", "RE", "RO", "RS", "RU",
    "RW", "SA", "SB", "SC", "SD", "SE", "SG", "SH", "SI", "SJ", "SK", "SL", "SM",
    "SN", "SO", "SR", "SS", "ST", "SV", "SX", "SY", "SZ", "TC", "TD",
    "TF", "TG", "TH", "TJ", "TK", "TL", "TM", "TN", "TO", "TR", "TT", "TV", "TW",
	"TZ", "UA", "UG", "UM", "US", "UY", "UZ", "VA",
    "VC", "VE", "VG", "VI", "VN", "VU", "WF", "WS", "YE", "YT", "ZA",
    "ZM", "ZW", "AN", "AP", "EU", "??",
};
//---------------------------------------------------------------------------

IP2Country::IP2Country() {
    hFlagImages = ImageList_LoadImage(g_hInstance, MAKEINTRESOURCE(IDB_FLAGS), 25, 4, CLR_DEFAULT, IMAGE_BITMAP, LR_CREATEDIBSECTION | LR_SHARED);

    ui32Count = 0;
    ui32Size = 65536;

	ui32RangeFrom = NULL;
    ui32RangeTo = NULL;
    ui8RangeCI = NULL;

    ui32IPv6Count = 0;
    ui32IPv6Size = 65536;

	ui128IPv6RangeFrom = NULL;
    ui128IPv6RangeTo = NULL;
    ui8IPv6RangeCI = NULL;

	FILE * ip2country = fopen((Util::getConfigPath() + "IpToCountry.csv").c_str(), "r");

    if(ip2country == NULL) {
        return;
    }

	ui32RangeFrom = (uint32_t *)calloc(ui32Size, sizeof(uint32_t));

    if(ui32RangeFrom == NULL) {
		fclose(ip2country);
		return;
    }

	ui32RangeTo = (uint32_t *)calloc(ui32Size, sizeof(uint32_t));

    if(ui32RangeTo == NULL) {
		fclose(ip2country);
		return;
    }

	ui8RangeCI = (uint8_t *)calloc(ui32Size, sizeof(uint8_t));

    if(ui8RangeCI == NULL) {
		fclose(ip2country);
		return;
	}

    char sLine[1024];

    while(fgets(sLine, 1024, ip2country) != NULL) {
        if(sLine[0] != '\"') {
            continue;
        }

        if(ui32Count == ui32Size) {
            ui32Size += 512;
            void * oldbuf = ui32RangeFrom;

			ui32RangeFrom = (uint32_t *)realloc(oldbuf, ui32Size * sizeof(uint32_t));

            if(ui32RangeFrom == NULL) {
                free(oldbuf);
    			fclose(ip2country);
                return;
    		}

            oldbuf = ui32RangeTo;

			ui32RangeTo = (uint32_t *)realloc(oldbuf, ui32Size * sizeof(uint32_t));

            if(ui32RangeTo == NULL) {
                free(oldbuf);
    			fclose(ip2country);
                return;
    		}

            oldbuf = ui8RangeCI;

			ui8RangeCI = (uint8_t *)realloc(oldbuf, ui32Size * sizeof(uint8_t));

            if(ui8RangeCI == NULL) {
                free(oldbuf);
    			fclose(ip2country);
                return;
    		}
        }

        char * sStart = sLine+1;
        uint8_t ui8d = 0;
        
        size_t iLineLen = strlen(sLine);

		for(size_t ui16i = 1; ui16i < iLineLen; ui16i++) {
            if(sLine[ui16i] == '\"') {
                sLine[ui16i] = '\0';
                if(ui8d == 0) {
                    ui32RangeFrom[ui32Count] = strtoul(sStart, NULL, 10);
                } else if(ui8d == 1) {
                    ui32RangeTo[ui32Count] = strtoul(sStart, NULL, 10);
                } else if(ui8d == 4) {
                    for(uint8_t ui8i = 0; ui8i < 252; ui8i++) {
                        if(*((uint16_t *)CountryCodes[ui8i]) == *((uint16_t *)sStart)) {
                            ui8RangeCI[ui32Count] = ui8i;
                            ui32Count++;
                            break;
                        }
                    }

                    break;
                }

                ui8d++;
                ui16i += (uint16_t)2;
                sStart = sLine+ui16i+1;
                
            }
        }
    }

	fclose(ip2country);

	ip2country = fopen((Util::getConfigPath() + "IpToCountry.6R.csv").c_str(), "r");

    if(ip2country == NULL) {
        return;
    }

	ui128IPv6RangeFrom = (uint8_t *)calloc(ui32IPv6Size, sizeof(uint8_t) * 16);

    if(ui128IPv6RangeFrom == NULL) {
		fclose(ip2country);
		return;
    }

	ui128IPv6RangeTo = (uint8_t *) calloc(ui32IPv6Size, sizeof(uint8_t) * 16);

    if(ui128IPv6RangeTo == NULL) {
		fclose(ip2country);
		return;
    }

	ui8IPv6RangeCI = (uint8_t *)calloc(ui32IPv6Size, sizeof(uint8_t));

    if(ui8IPv6RangeCI == NULL) {
		fclose(ip2country);
		return;
	}

    while(fgets(sLine, 1024, ip2country) != NULL) {
        if(sLine[0] == '#' || sLine[0] < 32) {
            continue;
        }

        if(ui32IPv6Count == ui32IPv6Size) {
            ui32IPv6Size += 512;
            void * oldbuf = ui128IPv6RangeFrom;

			ui128IPv6RangeFrom = (uint8_t *) realloc(oldbuf, ui32IPv6Size * (sizeof(uint8_t)*16));

            if(ui128IPv6RangeFrom == NULL) {
                free(oldbuf);
    			fclose(ip2country);
                return;
    		}

            oldbuf = ui128IPv6RangeTo;

			ui128IPv6RangeTo = (uint8_t *) realloc(oldbuf, ui32IPv6Size * (sizeof(uint8_t)*16));

            if(ui128IPv6RangeTo == NULL) {
                free(oldbuf);
    			fclose(ip2country);
                return;
    		}

            oldbuf = ui8IPv6RangeCI;

			ui8IPv6RangeCI = (uint8_t *) realloc(oldbuf, ui32IPv6Size * sizeof(uint8_t));

            if(ui8IPv6RangeCI == NULL) {
                free(oldbuf);
    			fclose(ip2country);
                return;
    		}
        }

        char * sStart = sLine;
        uint8_t ui8d = 0;

        size_t iLineLen = strlen(sLine);

		for(size_t ui16i = 0; ui16i < iLineLen; ui16i++) {
            if(ui8d == 0 && sLine[ui16i] == '-') {
                sLine[ui16i] = '\0';

                Socket::inet_pton(sStart, ui128IPv6RangeFrom + (ui32IPv6Count*16));
            } else if(sLine[ui16i] == ',') {
                sLine[ui16i] = '\0';
                if(ui8d == 1) {
                    Socket::inet_pton(sStart, ui128IPv6RangeTo + (ui32IPv6Count*16));
                } else {
                    for(uint8_t ui8i = 0; ui8i < 252; ui8i++) {
                        if(*((uint16_t *)CountryCodes[ui8i]) == *((uint16_t *)sStart)) {
                            ui8IPv6RangeCI[ui32IPv6Count] = ui8i;
                            ui32IPv6Count++;

                            break;
                        }
                    }

                    break;
                }
            } else {
                continue;
            }

            ui8d++;
            sStart = sLine+ui16i+1;
        }
    }

	fclose(ip2country);
}
//---------------------------------------------------------------------------
	
IP2Country::~IP2Country() {
    ImageList_Destroy(hFlagImages);

	free(ui32RangeFrom);
	free(ui32RangeTo);
	free(ui8RangeCI);

	free(ui128IPv6RangeFrom);
	free(ui128IPv6RangeTo);
	free(ui8IPv6RangeCI);
}
//---------------------------------------------------------------------------

string IP2Country::GetCountryFromIP(const string &sIP) {
	if(BOOLSETTING(GET_USER_COUNTRY) == false) {
        return Util::emptyString;
    }

    if(strchr(sIP.c_str(), '.') != NULL) {
        uint32_t ui32IP = inet_addr(sIP.c_str());

        uint32_t ui32IpHash = 0;
        ui32IpHash = 16777216 * ((uint8_t *)&ui32IP)[0] + 65536 * ((uint8_t *)&ui32IP)[1] + 256 * ((uint8_t *)&ui32IP)[2] + ((uint8_t *)&ui32IP)[3];

        for(uint32_t ui32i = 0; ui32i < ui32Count; ui32i++) {
            if(ui32RangeFrom[ui32i] < ui32IpHash && ui32RangeTo[ui32i] > ui32IpHash) {
                return CountryCodes[ui8RangeCI[ui32i]];
            }
        }
    } else {
        char ui128Ip[16];
        if(Socket::inet_pton(sIP.c_str(), ui128Ip) != 1) {
            return Util::emptyString;
		}

        for(uint32_t ui32i = 0; ui32i < ui32IPv6Count; ui32i++) {
            if(memcmp(ui128IPv6RangeFrom+(ui32i*16), ui128Ip, 16) <= 0 && memcmp(ui128IPv6RangeTo+(ui32i*16), ui128Ip, 16) >= 0) {
                return CountryCodes[ui8IPv6RangeCI[ui32i]];
            }
        }
    }

	return Util::emptyString;
}
//----------------------------------------------------------------------------------------------------------------------------------------------------

uint8_t IP2Country::GetFlagImageIndex(const char * country, bool fullname/* = false*/) {
    uint8_t ui8Count = (uint8_t)(sizeof(CountryCodes) / sizeof(CountryCodes[0]));

	if(fullname == true) {
		for(uint8_t i = 0; i < ui8Count; i++) {
			if(tolower(country[0]) == tolower(CountryNames[i][0]) && _stricmp(country+1, CountryNames[i]+1) == 0) {
				return i + 1;
			}
		}
	} else {
		for(uint8_t i = 0; i < ui8Count; i++) {
            if((*(uint16_t *)country) == (*(uint16_t *)CountryCodes[i])) {
				return i + 1;
			}
		}
	}

	return 0;
}
//----------------------------------------------------------------------------------------------------------------------------------------------------

uint8_t IP2Country::GetFlagImageIndexFromIP(const string &sIP) {
    if(strchr(sIP.c_str(), '.') != NULL) {
        uint32_t ui32IP = inet_addr(sIP.c_str());

        uint32_t ui32IpHash = 0;
        ui32IpHash = 16777216 * ((uint8_t *)&ui32IP)[0] + 65536 * ((uint8_t *)&ui32IP)[1] + 256 * ((uint8_t *)&ui32IP)[2] + ((uint8_t *)&ui32IP)[3];

        for(uint32_t ui32i = 0; ui32i < ui32Count; ui32i++) {
            if(ui32RangeFrom[ui32i] < ui32IpHash && ui32RangeTo[ui32i] > ui32IpHash) {
                return ui8RangeCI[ui32i] + 1;
            }
        }
    } else {
		char ui128Ip[16];
		if(Socket::inet_pton(sIP.c_str(), ui128Ip) != 1) {
            return 0;
		}

        for(uint32_t ui32i = 0; ui32i < ui32IPv6Count; ui32i++) {
            if(memcmp(ui128IPv6RangeFrom+(ui32i*16), ui128Ip, 16) <= 0 &&
                memcmp(ui128IPv6RangeTo+(ui32i*16), ui128Ip, 16) >= 0) {
                return ui8IPv6RangeCI[ui32i] + 1;
            }
        }
    }

    return 0;
}
//----------------------------------------------------------------------------------------------------------------------------------------------------
